from datetime import datetime
import json
import httplib

from mock import patch

from blockerbugs import db
from blockerbugs import app
from blockerbugs.models.spin import Spin, SpinState, SpinType
from blockerbugs.models.build import Build
from blockerbugs.models.userinfo import UserInfo
from testing.test_controllers import add_release, add_milestone, \
    add_bug, add_spin, add_update
from blockerbugs.controllers.api import errors


class TestRestAPI(object):
    @classmethod
    def setup_class(cls):
        db.session.rollback()
        db.drop_all()
        db.create_all()
        cls.client = app.test_client()
        cls.release = add_release(99)
        cls.milestone = add_milestone(cls.release, 'final', 100, 101,
                                      '99-final', True)
        cls.milestone2 = add_milestone(cls.release, 'beta', 200, 201,
                                       '99-beta')
        bug1 = add_bug(9000, 'testbug1', cls.milestone)
        bug1.accepted_fe = True
        bug2 = add_bug(9002, 'testbug2', cls.milestone)
        bug2.accepted_blocker = False
        bug2.proposed_fe = True
        cls.update_pending_stable = add_update(u'test-pending-stable.fc99',
                                               u'stable', [bug1], cls.release,
                                               [cls.milestone, cls.milestone2])
        cls.update_pending_stable.date_submitted = datetime(1990, 1, 1)
        cls.update_pending_stable.pending = True
        cls.update_testing2 = add_update(u'test-testing2.fc99', u'testing',
                                         [bug2],
                                         cls.release, [cls.milestone])
        cls.spin2 = add_spin('test-spin2', SpinType.TC, [],
                             cls.milestone, state=SpinState.requested)

        build = Build()
        build.koji_id = 33
        build.nvr = 'libofx-0.9.9-1.fc20'
        cls.spin1 = add_spin('test-spin1', SpinType.TC,
                             [cls.update_pending_stable],
                             cls.milestone, state=SpinState.requested)
        cls.spin1.builds = [build]
        cls.spin1.date_requested = datetime(1990, 1, 1)
        cls.spin1.succeeds = cls.spin2
        cls.spin2.date_requested = datetime(2000, 1, 1)
        cls.api_user = UserInfo('api_user')
        passwd = cls.api_user.generate_api_key()
        cls.headers = {'X-Auth-User': 'api_user', 'X-Auth-Key': passwd}
        db.session.add(cls.api_user)
        db.session.commit()

    @classmethod
    def teardown_class(cls):
        db.session.rollback()
        db.drop_all()

    def test_unknown_release(self):
        url = '/api/v0/milestones/9999/unknown/bugs'
        resp = self.client.get(url)
        assert resp.status_code == httplib.NOT_FOUND
        error = json.loads(resp.data)['error']
        assert error['code'] == errors.NoSuchObjectError.code
        assert 'Release' in error['message']

    def test_unknown_milestone(self):
        url = '/api/v0/milestones/99/unknown/bugs'
        resp = self.client.get(url)
        assert resp.status_code == httplib.NOT_FOUND
        error = json.loads(resp.data)['error']
        assert error['code'] == errors.NoSuchObjectError.code
        assert 'Milestone' in error['message']

    def test_list_all_bugs(self):
        url = '/api/v0/milestones/99/final/bugs'
        resp = self.client.get(url)
        assert resp.status_code == httplib.OK
        data = json.loads(resp.data)
        assert len(data) == 2
        bug = data[0]
        assert bug['bugid'] == 9000
        assert bug['url'] == 'https://bugzilla.redhat.com/show_bug.cgi?id=9000'
        assert bug['summary'] == 'testbug1'
        assert bug['component'] == 'testcomponent'
        assert bug['active']
        assert set(bug['type']) == set(('accepted_blocker', 'accepted_fe'))

    def test_list_accepted_blocker_bugs(self):
        url = '/api/v0/milestones/99/final/bugs?bugtype=accepted_blocker&'
        resp = self.client.get(url)
        assert resp.status_code == httplib.OK
        data = json.loads(resp.data)
        assert len(data) == 1
        assert data[0]['bugid'] == 9000

    def test_list_all_updates(self):
        url = '/api/v0/milestones/99/final/updates'
        resp = self.client.get(url)
        assert resp.status_code == httplib.OK
        data = json.loads(resp.data)
        assert len(data) == 2
        update = data[-1]
        assert update['title'] == u'test-pending-stable.fc99'
        assert update['url'] == u'http://localhost/update'
        assert update['karma'] == 1
        assert update['stable_karma'] == 3
        assert update['status'] == u'stable'
        assert update['pending']
        assert update['bugs'][0]['bugid'] == 9000
        assert set(update['bugs'][0]['type']) == set(('accepted_blocker',
                                                  'accepted_fe'))
        assert update['release'] == 99
        assert len(update['milestones']) == 2
        assert {'version': 'final', 'release': 99} in update['milestones']
        assert {'version': 'beta', 'release': 99} in update['milestones']

    def test_list_updates_with_proposed_fe_bugs(self):
        url = '/api/v0/milestones/99/final/updates?bugtype=proposed_fe&'
        resp = self.client.get(url)
        assert resp.status_code == httplib.OK
        data = json.loads(resp.data)
        assert len(data) == 1
        update = data[0]
        assert update['title'] == u'test-testing2.fc99'

    def test_list_spins(self):
        url = '/api/v0/milestones/99/final/spins'
        resp = self.client.get(url)
        assert resp.status_code == httplib.OK
        data = json.loads(resp.data)
        assert len(data) == 2
        spin = Spin.query.filter_by(name='test-spin1').first()
        resp_spin = data[1]
        assert resp_spin['name'] == spin.name
        assert resp_spin['id'] == spin.id
        assert resp_spin['state'] == spin.state.value
        assert resp_spin['url'] == spin.url
        assert resp_spin['date_requested'] == datetime(1990, 1, 1).isoformat()
        assert resp_spin['date_created'] is None
        assert resp_spin['spin_type'] == spin.spin_type.value
        assert len(resp_spin['updates']) == 1
        assert {'title': u'test-pending-stable.fc99',
                'status': u'stable',
                'url': u'http://localhost/update'} in resp_spin['updates']
        succeeded = Spin.query.filter_by(name='test-spin2').first()
        assert resp_spin['succeeds']['name'] == succeeded.name
        assert resp_spin['succeeds']['id'] == succeeded.id
        assert {'version': 'final', 'release': 99} == resp_spin['milestone']

    def test_basic_create_spins(self):
        url = '/api/v0/milestones/99/final/spins'
        test_spin_data = {
            'name': 'create_spin_test1',
            'spin_type': 'TC',
            'updates': ['test-pending-stable.fc99'],
        }
        with patch('blockerbugs.models.spin.datetime') as mock_date:
            mock_date.utcnow.return_value = datetime(1990, 1, 1)
            mock_date.side_effect = lambda *args, **kw: datetime(*args, **kw)
            resp = self.client.post(url,
                                    data=json.dumps(test_spin_data),
                                    content_type='application/json',
                                    headers=self.headers)
        assert resp.status_code == httplib.CREATED
        assert 'Location' in resp.headers
        spin = Spin.query.filter_by(name='create_spin_test1').first()
        assert spin.name == 'create_spin_test1'
        assert spin.state == SpinState.requested
        assert spin.url is None
        assert spin.spin_type == SpinType.TC
        assert len(spin.updates) == 1
        assert spin.updates[0].title == 'test-pending-stable.fc99'
        assert spin.succeeds is None
        assert spin.milestone.version == 'final'
        assert spin.milestone.release.number == 99
        assert spin.date_requested == datetime(1990, 1, 1)

    def test_create_spin_fails_unknown_update_title(self):
        url = '/api/v0/milestones/99/final/spins'
        test_spin_data = {
            'name': 'create_spin_test1',
            'spin_type': 'TC',
            'updates': ['test-pending-stable.fc99', 'unknown'],
        }
        resp = self.client.post(url,
                                data=json.dumps(test_spin_data),
                                content_type='application/json',
                                headers=self.headers)
        assert resp.status_code == httplib.NOT_FOUND
        error = json.loads(resp.data)['error']
        assert error['code'] == errors.NoSuchObjectError.code
        assert 'Update' in error['message']

    def test_invalid_spin_name(self):
        url = '/api/v0/milestones/99/final/spins'
        test_spin_data = {
            'name': 123,
            'spin_type': 'create_test1',
            'updates': ['test-pending-stable.fc99', 'unknown'],
            'succeeds': []
        }
        resp = self.client.post(url,
                                data=json.dumps(test_spin_data),
                                content_type='application/json',
                                headers=self.headers)
        assert resp.status_code == httplib.BAD_REQUEST
        error = json.loads(resp.data)['error']
        assert error['code'] == errors.ValidationError.code
        assert 'name' in error['message']

    def test_malformed_spin_data_request(self):
        url = '/api/v0/milestones/99/final/spins'
        test_spin_data = "{'name'= 123}"
        resp = self.client.post(url,
                                data=test_spin_data,
                                content_type='application/json',
                                headers=self.headers)
        assert resp.status_code == httplib.BAD_REQUEST
        error = json.loads(resp.data)['error']
        assert error['code'] == 1003

    def test_bad_bugtype_list_bugs(self):
        url = '/api/v0/milestones/99/final/updates?bugtype=foo&'
        resp = self.client.get(url)
        assert resp.status_code == httplib.BAD_REQUEST
        error = json.loads(resp.data)['error']
        assert error['code'] == errors.InvalidArgumentError.code
        assert 'bugtype' in error['message']

    def test_get_specific_spin_info(self):
        spin = Spin.query.filter_by(name='test-spin1').first()
        url = '/api/v0/milestones/99/final/spins/%d' % (spin.id)
        resp = self.client.get(url)
        assert resp.status_code == httplib.OK
        resp_spin = json.loads(resp.data)
        assert resp_spin['name'] == spin.name
        assert resp_spin['id'] == spin.id
        assert resp_spin['state'] == spin.state.value
        assert resp_spin['url'] == spin.url
        assert resp_spin['date_requested'] == datetime(1990, 1, 1).isoformat()
        assert resp_spin['date_created'] is None
        assert resp_spin['spin_type'] == spin.spin_type.value
        assert len(resp_spin['builds']) == 1
        build = resp_spin['builds'][0]
        assert build['id'] == spin.builds[0].koji_id
        assert build['nvr'] == spin.builds[0].nvr
        assert build['epoch'] == spin.builds[0].epoch

    def test_update_spin_created_date(self):
        spin_id = Spin.query.filter_by(name='test-spin2').first().id
        url = '/api/v0/milestones/99/final/spins/%d' % (spin_id)
        date_created = datetime(1991, 1, 1)
        update_data = {'date_created': date_created.isoformat(), }
        resp = self.client.put(url, data=json.dumps(update_data),
                               content_type='application/json',
                               headers=self.headers)
        assert resp.status_code == httplib.OK
        updated_spin = Spin.query.get(spin_id)
        assert updated_spin.date_created == date_created

    def test_update_spin_invalid_created_date(self):
        spin_id = Spin.query.filter_by(name='test-spin2').first().id
        url = '/api/v0/milestones/99/final/spins/%d' % (spin_id)
        date_created = 'invalid'
        update_data = {'date_created': date_created, }
        resp = self.client.put(url, data=json.dumps(update_data),
                               content_type='application/json',
                               headers=self.headers)
        assert resp.status_code == httplib.BAD_REQUEST
        error = json.loads(resp.data)['error']
        assert error['code'] == errors.ValidationError.code

    def test_update_spin_auth_failed(self):
        spin_id = Spin.query.filter_by(name='test-spin2').first().id
        url = '/api/v0/milestones/99/final/spins/%d' % (spin_id)
        update_data = {'date_created': 'foo', }
        resp = self.client.put(url, data=json.dumps(update_data),
                               content_type='application/json')
        assert resp.status_code == httplib.FORBIDDEN
        error = json.loads(resp.data)['error']
        assert error['code'] == errors.AuthFailedError.code
